/**********************************************************************
 Copyright (c) 2006 Thomas Marti, Stefan Schmid and others. All rights reserved.
 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.

 Contributors:
 2006 Erik Bengtson - Original SpatialExpression implementation
 ...
 **********************************************************************/
package org.datanucleus.store.expression.spatial;

import java.util.ArrayList;

import org.datanucleus.ClassLoaderResolver;
import org.datanucleus.store.mapped.expression.BinaryExpression;
import org.datanucleus.store.mapped.expression.CharacterLiteral;
import org.datanucleus.store.mapped.expression.QueryExpression;
import org.datanucleus.store.mapped.expression.ScalarExpression;
import org.datanucleus.store.mapped.expression.StringExpression;
import org.datanucleus.store.mapped.mapping.JavaTypeMapping;

/**
 * Allow different datastores to provide their own representation of spatial expresssions.
 */
public abstract class SpatialExpression extends ScalarExpression
{

    public SpatialExpression(QueryExpression qs)
    {
        super(qs);
    }

    // ---- OGC SF functions for constructing a geometry value given its well-known text representation ----//

    public ScalarExpression geomFromTextMethod(ScalarExpression wkt, ScalarExpression srid)
    {
        return geometryFunction("GeomFromText", wkt, srid, null);
    }

    public ScalarExpression pointFromTextMethod(ScalarExpression wkt, ScalarExpression srid)
    {
        return geometryFunction("PointFromText", wkt, srid, null);
    }

    public ScalarExpression lineFromTextMethod(ScalarExpression wkt, ScalarExpression srid)
    {
        return geometryFunction("LineFromText", wkt, srid, null);
    }

    public ScalarExpression polyFromTextMethod(ScalarExpression wkt, ScalarExpression srid)
    {
        return geometryFunction("PolyFromText", wkt, srid, null);
    }

    public ScalarExpression mPointFromTextMethod(ScalarExpression wkt, ScalarExpression srid)
    {
        return geometryFunction("MPointFromText", wkt, srid, null);
    }

    public ScalarExpression mLineFromTextMethod(ScalarExpression wkt, ScalarExpression srid)
    {
        return geometryFunction("MLineFromText", wkt, srid, null);
    }

    public ScalarExpression mPolyFromTextMethod(ScalarExpression wkt, ScalarExpression srid)
    {
        return geometryFunction("MPolyFromText", wkt, srid, null);
    }

    public ScalarExpression geomCollFromTextMethod(ScalarExpression wkt, ScalarExpression srid)
    {
        return geometryFunction("GeomCollFromText", wkt, srid, null);
    }

    // ---- OGC SF functions for constructing a geometric object given its well-known binary representation ----//

    public ScalarExpression geomFromWKBMethod(ScalarExpression wkb, ScalarExpression srid)
    {
        return geometryFunction("GeomFromWKB", wkb, srid, null);
    }

    public ScalarExpression pointFromWKBMethod(ScalarExpression wkb, ScalarExpression srid)
    {
        return geometryFunction("PointFromWKB", wkb, srid, null);
    }

    public ScalarExpression lineFromWKBMethod(ScalarExpression wkb, ScalarExpression srid)
    {
        return geometryFunction("LineFromWKB", wkb, srid, null);
    }

    public ScalarExpression polyFromWKBMethod(ScalarExpression wkb, ScalarExpression srid)
    {
        return geometryFunction("PolyFromWKB", wkb, srid, null);
    }

    public ScalarExpression mPointFromWKBMethod(ScalarExpression wkb, ScalarExpression srid)
    {
        return geometryFunction("MPointFromWKB", wkb, srid, null);
    }

    public ScalarExpression mLineFromWKBMethod(ScalarExpression wkb, ScalarExpression srid)
    {
        return geometryFunction("MLineFromWKB", wkb, srid, null);
    }

    public ScalarExpression mPolyFromWKBMethod(ScalarExpression wkb, ScalarExpression srid)
    {
        return geometryFunction("MPolyFromWKB", wkb, srid, null);
    }

    public ScalarExpression geomCollFromWKBMethod(ScalarExpression wkb, ScalarExpression srid)
    {
        return geometryFunction("GeomCollFromWKB", wkb, srid, null);
    }

    // ---- OGC SF basic methods on geometric objects ----//

    public ScalarExpression dimensionMethod(ScalarExpression geom)
    {
        JavaTypeMapping integerMapping = qs.getStoreManager().getMappingManager().getMappingWithDatastoreMapping(
            Integer.class, false, false, qs.getClassLoaderResolver());
        return numericFunction("Dimension", geom, integerMapping);
    }

    public ScalarExpression geometryTypeMethod(ScalarExpression geom)
    {
        return stringFunction("GeometryType", geom);
    }

    public ScalarExpression sridMethod(ScalarExpression geom)
    {
        JavaTypeMapping integerMapping = qs.getStoreManager().getMappingManager().getMappingWithDatastoreMapping(
            Integer.class, false, false, qs.getClassLoaderResolver());
        return numericFunction("SRID", geom, integerMapping);
    }

    public ScalarExpression envelopeMethod(ScalarExpression geom)
    {
        return geometryFunction("Envelope", geom, getGeomMapping(geom));
    }

    public ScalarExpression asTextMethod(ScalarExpression geom)
    {
        return stringFunction("AsText", geom);
    }

    public ScalarExpression asBinaryMethod(ScalarExpression geom)
    {
        ArrayList args = new ArrayList();
        args.add(geom);
        return new BinaryExpression("AsBinary", args);
    }

    public ScalarExpression isEmptyMethod(ScalarExpression geom)
    {
        return booleanFunction("IsEmpty", geom);
    }

    public ScalarExpression isSimpleMethod(ScalarExpression geom)
    {
        return booleanFunction("IsSimple", geom);
    }

    public ScalarExpression boundaryMethod(ScalarExpression geom)
    {
        return geometryFunction("Boundary", geom, getGeomMapping(geom));
    }

    // ---- OGC SF methods for testing spatial relations between geometric objects ----//

    public ScalarExpression equalsMethod(ScalarExpression geom1, ScalarExpression geom2)
    {
        return booleanFunction("Equals", geom1, geom2);
    }

    public ScalarExpression disjointMethod(ScalarExpression geom1, ScalarExpression geom2)
    {
        return booleanFunction("Disjoint", geom1, geom2);
    }

    public ScalarExpression intersectsMethod(ScalarExpression geom1, ScalarExpression geom2)
    {
        return booleanFunction("Intersects", geom1, geom2);
    }

    public ScalarExpression touchesMethod(ScalarExpression geom1, ScalarExpression geom2)
    {
        return booleanFunction("Touches", geom1, geom2);
    }

    public ScalarExpression crossesMethod(ScalarExpression geom1, ScalarExpression geom2)
    {
        return booleanFunction("Crosses", geom1, geom2);
    }

    public ScalarExpression withinMethod(ScalarExpression geom1, ScalarExpression geom2)
    {
        return booleanFunction("Within", geom1, geom2);
    }

    public ScalarExpression containsMethod(ScalarExpression geom1, ScalarExpression geom2)
    {
        return booleanFunction("Contains", geom1, geom2);
    }

    public ScalarExpression overlapsMethod(ScalarExpression geom1, ScalarExpression geom2)
    {
        return booleanFunction("Overlaps", geom1, geom2);
    }

    public ScalarExpression relateMethod(ScalarExpression geom1, ScalarExpression geom2, ScalarExpression patternMatrix)
    {
        return booleanFunction("Relate", geom1, geom2, patternMatrix);
    }

    // ---- OGC SF methods that support spatial analysis ----//

    public ScalarExpression distanceMethod(ScalarExpression geom1, ScalarExpression geom2)
    {
        JavaTypeMapping doubleMapping = qs.getStoreManager().getMappingManager().getMappingWithDatastoreMapping(
            Double.class, false, false, qs.getClassLoaderResolver());
        return numericFunction("Distance", geom1, geom2, doubleMapping);
    }

    public ScalarExpression bufferMethod(ScalarExpression geom, ScalarExpression distance)
    {
        return geometryFunction("Buffer", geom, distance, getGeomMapping(geom));
    }

    public ScalarExpression convexHullMethod(ScalarExpression geom)
    {
        return geometryFunction("ConvexHull", geom, getGeomMapping(geom));
    }

    public ScalarExpression intersectionMethod(ScalarExpression geom1, ScalarExpression geom2)
    {
        return geometryFunction("Intersection", geom1, geom2, getGeomMapping(geom1));
    }

    public ScalarExpression unionMethod(ScalarExpression geom1, ScalarExpression geom2)
    {
        return geometryFunction("Union", geom1, geom2, getGeomMapping(geom1));
    }

    public ScalarExpression differenceMethod(ScalarExpression geom1, ScalarExpression geom2)
    {
        return geometryFunction("Difference", geom1, geom2, getGeomMapping(geom1));
    }

    public ScalarExpression symDifferenceMethod(ScalarExpression geom1, ScalarExpression geom2)
    {
        return geometryFunction("SymDifference", geom1, geom2, getGeomMapping(geom1));
    }

    // ---- OGC SF methods on on type Point ----//

    public ScalarExpression xMethod(ScalarExpression geom)
    {
        JavaTypeMapping doubleMapping = qs.getStoreManager().getMappingManager().getMappingWithDatastoreMapping(
            Double.class, false, false, qs.getClassLoaderResolver());
        return numericFunction("X", geom, doubleMapping);
    }

    public ScalarExpression yMethod(ScalarExpression geom)
    {
        JavaTypeMapping doubleMapping = qs.getStoreManager().getMappingManager().getMappingWithDatastoreMapping(
            Double.class, false, false, qs.getClassLoaderResolver());
        return numericFunction("Y", geom, doubleMapping);
    }

    // ---- OGC SF methods on type Curve ----//

    public ScalarExpression startPointMethod(ScalarExpression geom)
    {
        return geometryFunction("StartPoint", geom, getGeomMapping(geom));
    }

    public ScalarExpression endPointMethod(ScalarExpression geom)
    {
        return geometryFunction("EndPoint", geom, getGeomMapping(geom));
    }

    public ScalarExpression isRingMethod(ScalarExpression geom)
    {
        return booleanFunction("IsRing", geom);
    }

    // ---- OGC SF methods on type Curve and MultiCurve----//

    public ScalarExpression isClosedMethod(ScalarExpression geom)
    {
        return booleanFunction("IsClosed", geom);
    }

    public ScalarExpression lengthMethod(ScalarExpression geom)
    {
        JavaTypeMapping doubleMapping = qs.getStoreManager().getMappingManager().getMappingWithDatastoreMapping(
            Double.class, false, false, qs.getClassLoaderResolver());
        return numericFunction("Length", geom, doubleMapping);
    }

    // ---- OGC SF methods on type LineString ----//

    public ScalarExpression numPointsMethod(ScalarExpression geom)
    {
        JavaTypeMapping integerMapping = qs.getStoreManager().getMappingManager().getMappingWithDatastoreMapping(
            Integer.class, false, false, qs.getClassLoaderResolver());
        return numericFunction("NumPoints", geom, integerMapping);
    }

    public ScalarExpression pointNMethod(ScalarExpression geom, ScalarExpression n)
    {
        return geometryFunction("PointN", geom, n, getGeomMapping(geom));
    }

    // ---- OGC SF methods on type Surface and MultiSurface ----//

    public ScalarExpression areaMethod(ScalarExpression geom)
    {
        JavaTypeMapping doubleMapping = qs.getStoreManager().getMappingManager().getMappingWithDatastoreMapping(
            Double.class, false, false, qs.getClassLoaderResolver());
        return numericFunction("Area", geom, doubleMapping);
    }

    public ScalarExpression centroidMethod(ScalarExpression geom)
    {
        return geometryFunction("Centroid", geom, getGeomMapping(geom));
    }

    public ScalarExpression pointOnSurfaceMethod(ScalarExpression geom)
    {
        return geometryFunction("PointOnSurface", geom, getGeomMapping(geom));
    }

    // ---- OGC SF methods on type Polygon ----//

    public ScalarExpression exteriorRingMethod(ScalarExpression geom)
    {
        return geometryFunction("ExteriorRing", geom, getGeomMapping(geom));
    }

    public ScalarExpression numInteriorRingMethod(ScalarExpression geom)
    {
        JavaTypeMapping integerMapping = qs.getStoreManager().getMappingManager().getMappingWithDatastoreMapping(
            Integer.class, false, false, qs.getClassLoaderResolver());
        return numericFunction("NumInteriorRing", geom, integerMapping);
    }

    public ScalarExpression interiorRingNMethod(ScalarExpression geom, ScalarExpression n)
    {
        return geometryFunction("InteriorRingN", geom, n, getGeomMapping(geom));
    }

    // ---- OGC SF methods on type GeomCollection ----//

    public ScalarExpression numGeometriesMethod(ScalarExpression geom)
    {
        JavaTypeMapping integerMapping = qs.getStoreManager().getMappingManager().getMappingWithDatastoreMapping(
            Integer.class, false, false, qs.getClassLoaderResolver());
        return numericFunction("NumGeometries", geom, integerMapping);
    }

    public ScalarExpression geometryNMethod(ScalarExpression geom, ScalarExpression n)
    {
        return geometryFunction("GeometryN", geom, n, getGeomMapping(geom));
    }

    // ---- bounding box ----//

    public abstract ScalarExpression bboxTestMethod(ScalarExpression geom1, ScalarExpression geom2);

    // ---- ----//

    protected ScalarExpression stringFunction(String functionName, ScalarExpression expr)
    {
        JavaTypeMapping stringMapping = qs.getStoreManager().getMappingManager().getMappingWithDatastoreMapping(
            String.class, false, false, qs.getClassLoaderResolver());
        ArrayList args = new ArrayList();
        args.add(expr);
        return new SpatialStringExpression(functionName, args, stringMapping);
    }

    protected ScalarExpression numericFunction(String functionName, ScalarExpression expr, JavaTypeMapping mapping)
    {
        ArrayList args = new ArrayList();
        args.add(expr);
        return new SpatialNumericExpression(functionName, args, mapping);
    }

    protected ScalarExpression numericFunction(String functionName, ScalarExpression expr1, ScalarExpression expr2, JavaTypeMapping mapping)
    {
        ArrayList args = new ArrayList();
        args.add(expr1);
        args.add(expr2);
        return new SpatialNumericExpression(functionName, args, mapping);
    }

    protected ScalarExpression numericFunction(String functionName, ScalarExpression expr1, ScalarExpression expr2, ScalarExpression expr3,
            JavaTypeMapping mapping)
    {
        ArrayList args = new ArrayList();
        args.add(expr1);
        args.add(expr2);
        args.add(expr3);
        return new SpatialNumericExpression(functionName, args, mapping);
    }

    protected ScalarExpression booleanFunction(String functionName, ScalarExpression expr)
    {
        ArrayList args = new ArrayList();
        args.add(expr);
        return new SpatialBooleanExpression(new StringExpression(functionName, args), new CharacterLiteral(qs, null, '1'));
    }

    protected ScalarExpression booleanFunction(String functionName, ScalarExpression expr1, ScalarExpression expr2)
    {
        ArrayList args = new ArrayList();
        args.add(expr1);
        args.add(expr2);
        return new SpatialBooleanExpression(new StringExpression(functionName, args), new CharacterLiteral(qs, null, '1'));
    }

    protected ScalarExpression booleanFunction(String functionName, ScalarExpression expr1, ScalarExpression expr2, ScalarExpression expr3)
    {
        ArrayList args = new ArrayList();
        args.add(expr1);
        args.add(expr2);
        args.add(expr3);
        return new SpatialBooleanExpression(new StringExpression(functionName, args), new CharacterLiteral(qs, null, '1'));
    }

    protected ScalarExpression geometryFunction(String functionName, ScalarExpression expr, JavaTypeMapping mapping)
    {
        ArrayList args = new ArrayList();
        args.add(expr);
        return new GeometryScalarExpression(functionName, args, mapping);
    }

    protected ScalarExpression geometryFunction(String functionName, ScalarExpression expr1, ScalarExpression expr2, JavaTypeMapping mapping)
    {
        ArrayList args = new ArrayList();
        args.add(expr1);
        args.add(expr2);
        return new GeometryScalarExpression(functionName, args, mapping);
    }

    protected ScalarExpression geometryFunction(String functionName, ScalarExpression expr1, ScalarExpression expr2,
            ScalarExpression expr3, JavaTypeMapping mapping)
    {
        ArrayList args = new ArrayList();
        args.add(expr1);
        args.add(expr2);
        args.add(expr3);
        return new GeometryScalarExpression(functionName, args, mapping);
    }

    protected ScalarExpression geometryFunction(String functionName, ScalarExpression expr1, ScalarExpression expr2,
            ScalarExpression expr3, ScalarExpression expr4, ScalarExpression expr5, JavaTypeMapping mapping)
    {
        ArrayList args = new ArrayList();
        args.add(expr1);
        args.add(expr2);
        args.add(expr3);
        args.add(expr4);
        args.add(expr5);
        return new GeometryScalarExpression(functionName, args, mapping);
    }

    protected ScalarExpression geometryFunction(String functionName, ScalarExpression expr, String asType, JavaTypeMapping mapping)
    {
        ArrayList args = new ArrayList();
        args.add(expr);
        ArrayList types = new ArrayList();
        types.add(asType);
        return new GeometryScalarExpression(functionName, args, types, mapping);
    }

    protected JavaTypeMapping getGeomMapping(ScalarExpression geom)
    {
        if (geom.getMapping() == null)
            return null;
        ClassLoaderResolver clr = qs.getClassLoaderResolver();
        String geometryPackageName = geom.getMapping().getJavaType().getPackage().getName();
        String className;
        if (geometryPackageName.equals("org.postgis"))
        {
            className = geometryPackageName + ".Geometry";
        }
        else if (geometryPackageName.equals("com.vividsolutions.jts.geom"))
        {
            className = geometryPackageName + ".Geometry";
        }
        else if (geometryPackageName.equals("oracle.spatial.geometry"))
        {
            className = geometryPackageName + ".JGeometry";
        }
        else
        {
            return null;
        }
        Class geometryClass = clr.classForName(className);
        if (geometryClass == null)
            return null;
        return qs.getStoreManager().getMappingManager().getMappingWithDatastoreMapping(
            geometryClass, false, false, clr);
    }
}