/**********************************************************************
 Copyright (c) 2006 Thomas Marti, Stefan Schmid and others. All rights reserved.
 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.

 Contributors:
 2006 Erik Bengtson - Original SpatialExpression implementation
 ...
 **********************************************************************/
package org.datanucleus.store.expression.spatial;

import java.lang.reflect.Constructor;
import java.util.HashMap;
import java.util.Map;

import org.datanucleus.jdo.spatial.SpatialHelper;
import org.datanucleus.ClassLoaderResolver;
import org.datanucleus.exceptions.NucleusException;
import org.datanucleus.plugin.ConfigurationElement;
import org.datanucleus.plugin.Extension;
import org.datanucleus.plugin.PluginManager;
import org.datanucleus.store.mapped.DatastoreAdapter;
import org.datanucleus.store.mapped.MappedStoreManager;
import org.datanucleus.store.mapped.expression.QueryExpression;
import org.datanucleus.store.mapped.expression.ScalarExpression;
import org.datanucleus.util.NucleusLogger;
import org.datanucleus.util.Localiser;

public class SpatialExpressionProxy extends SpatialExpression
{
    private static final Localiser LOCALISER = Localiser.getInstance("org.datanucleus.store.expression.spatial.Localisation",
        SpatialHelper.class.getClassLoader());

    private static Map spatialExpressionAdapters = new HashMap();

    private SpatialExpression currentAdapter;

    public SpatialExpressionProxy(QueryExpression qs)
    {
        super(qs);
        initializeExpressionSpatialAdapter();
    }

    private void initializeExpressionSpatialAdapter()
    {
        MappedStoreManager storeMgr = qs.getStoreManager();
        DatastoreAdapter datastoreAdapter = storeMgr.getDatastoreAdapter();
        ClassLoaderResolver clr = qs.getClassLoaderResolver();
        currentAdapter = (SpatialExpression) spatialExpressionAdapters.get(datastoreAdapter.getVendorID());
        if (currentAdapter == null)
        {
            PluginManager pluginMgr = storeMgr.getOMFContext().getPluginManager();
            Extension[] ex = pluginMgr.getExtensionPoint("org.datanucleus.spatial.spatial_adapter").getExtensions();
            for (int i = 0; currentAdapter == null && i < ex.length; i++)
            {
                ConfigurationElement[] confElm = ex[i].getConfigurationElements();
                for (int c = 0; currentAdapter == null && c < confElm.length; c++)
                {
                    if (confElm[c].getAttribute("vendor-id").equals(datastoreAdapter.getVendorID()))
                    {
                        ClassLoader cl = this.getClass().getClassLoader();
                        Class adapterClass = clr.classForName(confElm[c].getAttribute("class-name"), cl);
                        Constructor constructor = null;
                        try
                        {
                            constructor = adapterClass.getConstructor(new Class[]{QueryExpression.class});
                        }
                        catch (SecurityException e)
                        {
                            NucleusLogger.QUERY
                                    .warn(LOCALISER.msg("Spatial.constructorDoesntExist", confElm[c].getAttribute("class-name"), e));
                        }
                        catch (NoSuchMethodException e)
                        {
                            NucleusLogger.QUERY
                                    .warn(LOCALISER.msg("Spatial.constructorDoesntExist", confElm[c].getAttribute("class-name"), e));
                        }
                        try
                        {
                            if (constructor != null)
                            {
                                currentAdapter = (SpatialExpression) constructor.newInstance(new Object[]{qs});
                            }
                            else
                            {
                                currentAdapter = (SpatialExpression) adapterClass.newInstance();
                            }
                        }
                        catch (Exception e)
                        {
                            String msg = LOCALISER.msg("Spatial.newInstanceError", confElm[c].getAttribute("class-name"), e);
                            NucleusLogger.QUERY.error(msg, e);
                            throw new NucleusException(msg, e).setFatal();
                        }
                    }
                }
            }
            if (currentAdapter == null)
            {
                throw new NucleusException(LOCALISER.msg("Spatial.couldNotCreateAdapter", datastoreAdapter.getVendorID())).setFatal();
            }
            spatialExpressionAdapters.put(datastoreAdapter.getVendorID(), currentAdapter);
        }
    }

    // ---- OGC SF functions for constructing a geometry value given its well-known text representation ----//

    public ScalarExpression geomFromTextMethod(ScalarExpression wkt, ScalarExpression srid)
    {
        return currentAdapter.geomFromTextMethod(wkt, srid);
    }

    public ScalarExpression pointFromTextMethod(ScalarExpression wkt, ScalarExpression srid)
    {
        return currentAdapter.pointFromTextMethod(wkt, srid);
    }

    public ScalarExpression lineFromTextMethod(ScalarExpression wkt, ScalarExpression srid)
    {
        return currentAdapter.lineFromTextMethod(wkt, srid);
    }

    public ScalarExpression polyFromTextMethod(ScalarExpression wkt, ScalarExpression srid)
    {
        return currentAdapter.polyFromTextMethod(wkt, srid);
    }

    public ScalarExpression mPointFromTextMethod(ScalarExpression wkt, ScalarExpression srid)
    {
        return currentAdapter.mPointFromTextMethod(wkt, srid);
    }

    public ScalarExpression mLineFromTextMethod(ScalarExpression wkt, ScalarExpression srid)
    {
        return currentAdapter.mLineFromTextMethod(wkt, srid);
    }

    public ScalarExpression mPolyFromTextMethod(ScalarExpression wkt, ScalarExpression srid)
    {
        return currentAdapter.mPolyFromTextMethod(wkt, srid);
    }

    public ScalarExpression geomCollFromTextMethod(ScalarExpression wkt, ScalarExpression srid)
    {
        return currentAdapter.geomCollFromTextMethod(wkt, srid);
    }

    // ---- OGC SF functions for constructing a geometric object given its well-known binary representation ----//

    public ScalarExpression geomFromWKBMethod(ScalarExpression wkb, ScalarExpression srid)
    {
        return currentAdapter.geomFromWKBMethod(wkb, srid);
    }

    public ScalarExpression pointFromWKBMethod(ScalarExpression wkb, ScalarExpression srid)
    {
        return currentAdapter.pointFromWKBMethod(wkb, srid);
    }

    public ScalarExpression lineFromWKBMethod(ScalarExpression wkb, ScalarExpression srid)
    {
        return currentAdapter.lineFromWKBMethod(wkb, srid);
    }

    public ScalarExpression polyFromWKBMethod(ScalarExpression wkb, ScalarExpression srid)
    {
        return currentAdapter.polyFromWKBMethod(wkb, srid);
    }

    public ScalarExpression mPointFromWKBMethod(ScalarExpression wkb, ScalarExpression srid)
    {
        return currentAdapter.mPointFromWKBMethod(wkb, srid);
    }

    public ScalarExpression mLineFromWKBMethod(ScalarExpression wkb, ScalarExpression srid)
    {
        return currentAdapter.mLineFromWKBMethod(wkb, srid);
    }

    public ScalarExpression mPolyFromWKBMethod(ScalarExpression wkb, ScalarExpression srid)
    {
        return currentAdapter.mPolyFromWKBMethod(wkb, srid);
    }

    public ScalarExpression geomCollFromWKBMethod(ScalarExpression wkb, ScalarExpression srid)
    {
        return currentAdapter.geomCollFromWKBMethod(wkb, srid);
    }

    // ---- OGC SF basic methods on geometric objects ----//

    public ScalarExpression dimensionMethod(ScalarExpression geom)
    {
        return currentAdapter.dimensionMethod(geom);
    }

    public ScalarExpression geometryTypeMethod(ScalarExpression geom)
    {
        return currentAdapter.geometryTypeMethod(geom);
    }

    public ScalarExpression sridMethod(ScalarExpression geom)
    {
        return currentAdapter.sridMethod(geom);
    }

    public ScalarExpression envelopeMethod(ScalarExpression geom)
    {
        return currentAdapter.envelopeMethod(geom);
    }

    public ScalarExpression asTextMethod(ScalarExpression geom)
    {
        return currentAdapter.asTextMethod(geom);
    }

    public ScalarExpression asBinaryMethod(ScalarExpression geom)
    {
        return currentAdapter.asBinaryMethod(geom);
    }

    public ScalarExpression isEmptyMethod(ScalarExpression geom)
    {
        return currentAdapter.isEmptyMethod(geom);
    }

    public ScalarExpression isSimpleMethod(ScalarExpression geom)
    {
        return currentAdapter.isSimpleMethod(geom);
    }

    public ScalarExpression boundaryMethod(ScalarExpression geom)
    {
        return currentAdapter.boundaryMethod(geom);
    }

    // ---- OGC SF methods for testing spatial relations between geometric objects ----//

    public ScalarExpression equalsMethod(ScalarExpression geom1, ScalarExpression geom2)
    {
        return currentAdapter.equalsMethod(geom1, geom2);
    }

    public ScalarExpression disjointMethod(ScalarExpression geom1, ScalarExpression geom2)
    {
        return currentAdapter.disjointMethod(geom1, geom2);
    }

    public ScalarExpression intersectsMethod(ScalarExpression geom1, ScalarExpression geom2)
    {
        return currentAdapter.intersectsMethod(geom1, geom2);
    }

    public ScalarExpression touchesMethod(ScalarExpression geom1, ScalarExpression geom2)
    {
        return currentAdapter.touchesMethod(geom1, geom2);
    }

    public ScalarExpression crossesMethod(ScalarExpression geom1, ScalarExpression geom2)
    {
        return currentAdapter.crossesMethod(geom1, geom2);
    }

    public ScalarExpression withinMethod(ScalarExpression geom1, ScalarExpression geom2)
    {
        return currentAdapter.withinMethod(geom1, geom2);
    }

    public ScalarExpression containsMethod(ScalarExpression geom1, ScalarExpression geom2)
    {
        return currentAdapter.containsMethod(geom1, geom2);
    }

    public ScalarExpression overlapsMethod(ScalarExpression geom1, ScalarExpression geom2)
    {
        return currentAdapter.overlapsMethod(geom1, geom2);
    }

    public ScalarExpression relateMethod(ScalarExpression geom1, ScalarExpression geom2, ScalarExpression patternMatrix)
    {
        return currentAdapter.relateMethod(geom1, geom2, patternMatrix);
    }

    // ---- OGC SF methods that support spatial analysis ----//

    public ScalarExpression distanceMethod(ScalarExpression geom1, ScalarExpression geom2)
    {
        return currentAdapter.distanceMethod(geom1, geom2);
    }

    public ScalarExpression bufferMethod(ScalarExpression geom, ScalarExpression distance)
    {
        return currentAdapter.bufferMethod(geom, distance);
    }

    public ScalarExpression convexHullMethod(ScalarExpression geom)
    {
        return currentAdapter.convexHullMethod(geom);
    }

    public ScalarExpression intersectionMethod(ScalarExpression geom1, ScalarExpression geom2)
    {
        return currentAdapter.intersectionMethod(geom1, geom2);
    }

    public ScalarExpression unionMethod(ScalarExpression geom1, ScalarExpression geom2)
    {
        return currentAdapter.unionMethod(geom1, geom2);
    }

    public ScalarExpression differenceMethod(ScalarExpression geom1, ScalarExpression geom2)
    {
        return currentAdapter.differenceMethod(geom1, geom2);
    }

    public ScalarExpression symDifferenceMethod(ScalarExpression geom1, ScalarExpression geom2)
    {
        return currentAdapter.symDifferenceMethod(geom1, geom2);
    }

    // ---- OGC SF methods on on type Point ----//

    public ScalarExpression xMethod(ScalarExpression geom)
    {
        return currentAdapter.xMethod(geom);
    }

    public ScalarExpression yMethod(ScalarExpression geom)
    {
        return currentAdapter.yMethod(geom);
    }

    // ---- OGC SF methods on type Curve ----//

    public ScalarExpression startPointMethod(ScalarExpression geom)
    {
        return currentAdapter.startPointMethod(geom);
    }

    public ScalarExpression endPointMethod(ScalarExpression geom)
    {
        return currentAdapter.endPointMethod(geom);
    }

    public ScalarExpression isRingMethod(ScalarExpression geom)
    {
        return currentAdapter.isRingMethod(geom);
    }

    // ---- OGC SF methods on type Curve and MultiCurve----//

    public ScalarExpression isClosedMethod(ScalarExpression geom)
    {
        return currentAdapter.isClosedMethod(geom);
    }

    public ScalarExpression lengthMethod(ScalarExpression geom)
    {
        return currentAdapter.lengthMethod(geom);
    }

    // ---- OGC SF methods on type LineString ----//

    public ScalarExpression numPointsMethod(ScalarExpression geom)
    {
        return currentAdapter.numPointsMethod(geom);
    }

    public ScalarExpression pointNMethod(ScalarExpression geom, ScalarExpression n)
    {
        return currentAdapter.pointNMethod(geom, n);
    }

    // ---- OGC SF methods on type Surface and MultiSurface ----//

    public ScalarExpression areaMethod(ScalarExpression geom)
    {
        return currentAdapter.areaMethod(geom);
    }

    public ScalarExpression centroidMethod(ScalarExpression geom)
    {
        return currentAdapter.centroidMethod(geom);
    }

    public ScalarExpression pointOnSurfaceMethod(ScalarExpression geom)
    {
        return currentAdapter.pointOnSurfaceMethod(geom);
    }

    // ---- OGC SF methods on type Polygon ----//

    public ScalarExpression exteriorRingMethod(ScalarExpression geom)
    {
        return currentAdapter.exteriorRingMethod(geom);
    }

    public ScalarExpression numInteriorRingMethod(ScalarExpression geom)
    {
        return currentAdapter.numInteriorRingMethod(geom);
    }

    public ScalarExpression interiorRingNMethod(ScalarExpression geom, ScalarExpression n)
    {
        return currentAdapter.interiorRingNMethod(geom, n);
    }

    // ---- OGC SF methods on type GeomCollection ----//

    public ScalarExpression numGeometriesMethod(ScalarExpression geom)
    {
        return currentAdapter.numGeometriesMethod(geom);
    }

    public ScalarExpression geometryNMethod(ScalarExpression geom, ScalarExpression n)
    {
        return currentAdapter.geometryNMethod(geom, n);
    }

    // ---- bounding box ----//

    public ScalarExpression bboxTestMethod(ScalarExpression geom1, ScalarExpression geom2)
    {
        return currentAdapter.bboxTestMethod(geom1, geom2);
    }

}
