/**********************************************************************
 Copyright (c) 2006 Thomas Marti, Stefan Schmid and others. All rights reserved.
 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.

 Contributors:
 ...
 **********************************************************************/
package org.datanucleus.store.mapped.mapping.jts;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.ObjectOutputStream;

import org.datanucleus.metadata.AbstractMemberMetaData;
import org.datanucleus.metadata.MetaDataUtils;
import org.datanucleus.store.expression.spatial.GeometryLiteral;
import org.datanucleus.store.expression.spatial.GeometryScalarExpression;
import org.datanucleus.ClassLoaderResolver;
import org.datanucleus.ClassNameConstants;
import org.datanucleus.ObjectManager;
import org.datanucleus.StateManager;
import org.datanucleus.store.mapped.DatastoreContainerObject;
import org.datanucleus.store.mapped.MappedStoreManager;
import org.datanucleus.store.mapped.expression.LogicSetExpression;
import org.datanucleus.store.mapped.expression.QueryExpression;
import org.datanucleus.store.mapped.expression.ScalarExpression;
import org.datanucleus.store.mapped.mapping.DatastoreMapping;
import org.datanucleus.store.mapped.mapping.MappingCallbacks;
import org.datanucleus.store.mapped.mapping.SingleFieldMultiMapping;
import org.datanucleus.store.rdbms.mapping.RDBMSMapping;
import org.datanucleus.store.rdbms.mapping.oracle.OracleBlobRDBMSMapping;

import com.vividsolutions.jts.geom.Geometry;
import com.vividsolutions.jts.io.ParseException;
import com.vividsolutions.jts.io.WKTReader;

/**
 * Mapping for com.vividsolutions.jts.geom.Geometry to its datastore representation.
 */
public class GeometryMapping extends SingleFieldMultiMapping implements MappingCallbacks
{
    private static final Geometry sampleValue = createGeom("POINT(0 0)");

    private static WKTReader wktReader;

    private boolean mapUserdataObject = false;

    protected static final Geometry createGeom(String wkt)
    {
        if (wktReader == null)
        {
            wktReader = new WKTReader();
        }

        try
        {
            return wktReader.read(wkt);
        }
        catch (ParseException e)
        {
            throw new Error("AssertionError: Illegal sampleValue", e);
        }
    }

    public String getJavaTypeForDatastoreMapping(int index)
    {
        return getJavaType().getName();
    }

    /* (non-Javadoc)
     * @see org.datanucleus.store.mapped.mapping.JavaTypeMapping#initialize(org.datanucleus.store.mapped.MappedStoreManager, java.lang.String)
     */
    public void initialize(MappedStoreManager storeMgr, String type)
    {
        super.initialize(storeMgr, type);
        addDatastoreField(getJavaType().getName()); // Geometry
    }

    public void initialize(AbstractMemberMetaData mmd, DatastoreContainerObject container, ClassLoaderResolver clr)
    {
        super.initialize(mmd, container, clr);
        addDatastoreField(getJavaType().getName()); // Geometry

        String mappingExtension = MetaDataUtils.getValueForExtensionRecursively(mmd, "mapping");
        if (mappingExtension != null && mappingExtension.equalsIgnoreCase("no-userdata"))
        {
            mapUserdataObject = false;
        }
        else
        {
            mapUserdataObject = true;
            addDatastoreField(ClassNameConstants.JAVA_IO_SERIALIZABLE); // User Data Object
        }
    }

    public Class getJavaType()
    {
        return Geometry.class;
    }

    public Object getSampleValue(ClassLoaderResolver clr)
    {
        return sampleValue;
    }

    public Object getObject(ObjectManager om, Object datastoreResults, int[] exprIndex)
    {
        Geometry geom = (Geometry) getDatastoreMapping(0).getObject(datastoreResults, exprIndex[0]);

        if (geom == null)
        {
            return null;
        }

        if (mapUserdataObject)
        {
            Object userData = getDatastoreMapping(1).getObject(datastoreResults, exprIndex[1]);
            geom.setUserData(userData);
        }

        return geom;
    }

    public void setObject(ObjectManager om, Object datastoreStatement, int[] exprIndex, Object value)
    {
        Geometry geom = (Geometry) value;
        if (geom == null)
        {
            getDatastoreMapping(0).setObject(datastoreStatement, exprIndex[0], null);
            if (exprIndex.length == 2)
            {
                getDatastoreMapping(1).setObject(datastoreStatement, exprIndex[1], null);
            }
        }
        else
        {
            getDatastoreMapping(0).setObject(datastoreStatement, exprIndex[0], geom);
            if (exprIndex.length == 2)
            {
                DatastoreMapping mapping = getDatastoreMapping(1);
                if (mapping instanceof RDBMSMapping && !((RDBMSMapping)mapping).insertValuesOnInsert()) 
                {
                    
                } 
                else 
                {
                    getDatastoreMapping(1).setObject(datastoreStatement, exprIndex[1], geom.getUserData());
                }
            }
        }
    }

    public void insertPostProcessing(StateManager sm)
    {
	}

	/**
     * Some nasty stuff that's only needed for Oracle.
     */
    public void postInsert(StateManager sm)
    {
        if (!mapUserdataObject || !(getDatastoreMapping(1) instanceof OracleBlobRDBMSMapping)) return;
        
        Object geom = sm.provideField(mmd.getAbsoluteFieldNumber());
        if (geom == null || !(geom instanceof Geometry) || ((Geometry)geom).getUserData() == null)
        {
            return;
        }

        Object value = ((Geometry)geom).getUserData();
        
        byte[] bytes = new byte[0];
        try
        {
            ByteArrayOutputStream baos = new ByteArrayOutputStream();
            ObjectOutputStream oos = new ObjectOutputStream(baos);
            oos.writeObject(value);
            bytes = baos.toByteArray();
        }
        catch (IOException e1)
        {
            // Do Nothing
        }

        // Update the BLOB
        OracleBlobRDBMSMapping.updateBlobColumn(sm, getDatastoreContainer(), getDatastoreMapping(1), bytes);
    }
    
    public void postUpdate(StateManager sm)
    {
        postInsert(sm);
    }

    private GeometryMapping cloneMapping()
    {
        GeometryMapping geomMapping = new GeometryMapping();
        geomMapping.storeMgr = this.storeMgr;
        geomMapping.type = this.type;
        geomMapping.mmd = this.mmd;
        geomMapping.mapUserdataObject = this.mapUserdataObject;
        geomMapping.datastoreContainer = this.datastoreContainer;
        geomMapping.datastoreMappings = new DatastoreMapping[1];
        geomMapping.datastoreMappings[0] = this.datastoreMappings[0]; // Geometry
        return geomMapping;
    }

    public void deleteDependent(StateManager sm)
    {
        // do nothing
    }

    public void postFetch(StateManager sm)
    {
        // do nothing
    }

    public void preDelete(StateManager sm)
    {
        // do nothing
    }

    // TODO Drop these when we drop legacy querying
    public ScalarExpression newLiteral(QueryExpression qs, Object value)
    {
        GeometryMapping geomMapping = cloneMapping();
        return new GeometryLiteral(qs, geomMapping, value);
    }

    public ScalarExpression newScalarExpression(QueryExpression qs, LogicSetExpression te)
    {
        GeometryMapping geomMapping = cloneMapping();
        return new GeometryScalarExpression(qs, geomMapping, te);
    }
}